import os
from fastapi import APIRouter
from fastapi import Request

from ..models.project import project
from .dtos import Result
from .dtos import ProjectInfo

router = APIRouter(prefix='/project')


@router.get('', response_model=Result)
def get_project_info(request: Request) -> Result:
    proj = project(request)
    api_key = os.environ.get('API_KEY', None)

    info = ProjectInfo(
        project_name=proj.project_name,
        description=proj.description,
        bindings=proj.bindings,
        rules=proj.rules,
        instances=proj.instances,
        api_key=api_key
    )
    return Result(status='OK', data=info)



from __future__ import annotations
from pydantic import Field

from ..base import BaseSchema


class Mapping(BaseSchema):
    id: str
    group: str
    base: str | None
    name: str
    instances: list[str] = []
    description: str = ''
    deployment: bool

    # FIXME https://github.com/pydantic/pydantic/issues/5992
    parent: Mapping | None = Field(default=None)



import csv
from dataclasses import dataclass
from dataclasses import field

from ..models.schema import Schema
from ..models.schema import Table
from ..models.schema import Field
from ..models.schema import Index


@dataclass
class Entity:
    instance: str = None
    table_name: str = None
    display_name: str = None
    fields: list = field(default_factory=list)
    indexes: list = field(default_factory=list)


@dataclass
class Relation:
    entity1: str = None
    entity2: str = None
    fields1: str = None
    fields2: str = None


class Parser:

    def __init__(self, mapping):
        self.mode = None
        self.mapping = mapping
        self.instances = {}
        self.relations = {}
        self.session = None

    def parse_line(self, line_no, line):
        if line == '':
            if self.mode == 2:
                if self.session.instance is not None:
                    if self.session.instance not in self.instances:
                        self.instances[self.session.instance] = []
                    self.instances[self.session.instance].append(self.session)
            if self.mode == 3:
                if self.session.entity2 not in self.relations:
                    self.relations[self.session.entity2] = {}
                for fld in self.session.fields2.split(','):
                    self.relations[self.session.entity2][fld] = self.session
            # 空行の場合、セッション解除
            self.mode = None
            return
        if line[0] == '#':
            return
        if self.mode is not None:
            name, value = self.split(line)
            if self.mode == 1:
                # 何もしない
                pass
            elif self.mode == 2:
                self.parse_entity(name, value)
            elif self.mode == 3:
                self.parse_relation(name, value)
            elif self.mode == 4:
                # コメントなので何もしない
                pass
            else:
                raise RuntimeError(f'Unknown Mode: {self.mode}')
        else:
            if line == '[Manager]':
                self.mode = 1
            elif line == '[Entity]':
                self.mode = 2
                self.session = Entity()
            elif line == '[Relation]':
                self.mode = 3
                self.session = Relation()
            elif line == '[Comment]':
                self.mode = 4
            else:
                raise RuntimeError(f'Unknown Line: ({line_no}) {line}')

    def split(self, line):
        work = line.split('=')
        if work[0] == "Index":
            # Indexは==となっている。
            # FIXME 恐らくIndex名を指定した場合を考慮する必要あり
            return work[0], work[2]
        return work[0], work[1]

    def parse_entity(self, name, value):
        if name == 'PName':
            self.session.table_name = value
        if name == 'LName':
            self.session.display_name = value
        if name == 'Page':
            value = value.upper()
            self.session.instance = self.mapping[value] if value in self.mapping else None
        if name == "Field":
            self.session.fields.append(value)
        if name == "Index":
            self.session.indexes.append(value)

    def parse_relation(self, name, value):
        if name == "Entity1":
            self.session.entity1 = value
        if name == "Entity2":
            self.session.entity2 = value
        if name == "Fields1":
            self.session.fields1 = value
        if name == "Fields2":
            self.session.fields2 = value


def convert_to_schema(p):
    schemas = {k: Schema(k) for k in p.instances.keys()}

    for key, entities in p.instances.items():
        for entity in entities:
            tbl = Table(
                instance=key,
                table_name=entity.table_name,
                display_name=entity.display_name
            )
            schemas[key].add_table(tbl)
            relation = p.relations[entity.table_name] if entity.table_name in p.relations else {}
            for row in csv.reader(entity.fields):
                # FIXME 恐らくdefault_valueの空文字初期化はうまく設定できない。
                field = Field(
                    display_name=row[0],
                    column_name=row[1],
                    column_type=row[2],
                    nullable=row[3] != 'NOT NULL',
                    primary_key=int(row[4]) if row[4] != '' else None,
                    default_value=row[5] if row[5] != '' else None,
                    foreign_key=relation[row[1]].entity1 if row[1] in relation else None,
                    comment=row[6] if row[6] != '' else None
                )
                tbl.fields.append(field)

            for row in csv.reader(entity.indexes):
                idx = Index(
                    index_name=None,
                    columns=row[1:]
                )
                tbl.indexes.append(idx)

    return schemas


def retrieve(folder, filename, mapping, **kwargs):
    p = Parser(mapping)
    with open(f'{folder}/{filename}', 'r', encoding='utf-8-sig') as f:
        for idx, line in enumerate(f):
            p.parse_line(idx + 1, line.strip())

    return convert_to_schema(p)



from . import engine

from ..models.schema import Table
from ..models.schema import Field


def is_exist(conn, env: str, table: Table):
    result = engine.select_one(
        conn,
        '''
        SELECT TABLE_NAME FROM information_schema.tables
        WHERE table_schema = :env and table_name = :table
        ''',
        {'env': env, 'table': table.table_name})
    return result is not None


def drop(conn, env: str, table: Table):
    sql = f'DROP TABLE {env}.{table.table_name}'
    engine.execute(conn, sql)


def _column_sql(field: dict):
    sql = f'`{field.column_name}` {field.column_type}'
    if not field.nullable:
        sql += ' NOT NULL'
    if field.default_value is not None:
        sql += f' DEFAULT {field.default_value}'
    return sql


def create(conn, env: str, table: Table):
    pk = sorted(
        [f for f in table.fields if f.primary_key is not None],
        key=lambda x: x.primary_key
    )
    sql = f'CREATE TABLE {env}.{table.table_name} ('
    sql += ', '.join([_column_sql(f) for f in table.fields])
    sql += f', constraint {table.table_name}_PKC primary key '
    sql += '(' + ', '.join([f'`{f.column_name}`' for f in pk]) + ')'
    sql += ')'
    engine.execute(conn, sql)

    for idx, index in enumerate(table.indexes):
        sql = 'CREATE INDEX '
        sql += f'{table.table_name}_IX{idx} ' if index.index_name is None else f'{index.index_name} '
        sql += f'ON {env}.{table.table_name} ('
        sql += ', '.join(index.columns)
        sql += ')'
        engine.execute(conn, sql)


def _col_value(item: dict, field: Field):
    if field.column_name in item:
        if item[field.column_name] is not None and type(item[field.column_name]) is str:
            # 関数定義されている場合には、SQLに関数を埋め込む。
            if '(' in item[field.column_name]:
                return item[field.column_name]
        return f':{field.column_name}'
    return 'NULL'


def insert(conn, env: str, table: Table, items: list[dict]):
    sql = f'INSERT INTO {env}.{table.table_name} ('
    sql += ', '.join([f'`{f.column_name}`' for f in table.fields])
    sql += ') VALUES ('
    sql += ', '.join([_col_value(items[0], f) for f in table.fields])
    sql += ')'
    engine.execute(conn, sql, items)
    conn.commit()


def backup(conn, env: str, table: Table, ymd: str):
    sql = f'CREATE TABLE {env}.bak_{table.table_name}_{ymd} AS SELECT * FROM {env}.{table.table_name}'
    engine.execute(conn, sql)


def restore(conn, env: str, table: Table, ymd: str):
    sql = f'INSERT IGNORE INTO {env}.{table.table_name} SELECT * FROM {env}.bak_{table.table_name}_{ymd}'
    engine.execute(conn, sql)


def is_exist_backup(conn, env: str, table: Table, ymd: str):
    result = engine.select_one(
        conn,
        '''
        SELECT TABLE_NAME FROM information_schema.tables
        WHERE table_schema = :env and table_name = :table
        ''',
        {'env': env, 'table': f'bak_{table.table_name}_{ymd}'})
    return result is not None



import logging
from argparse import ArgumentParser

from .models.project import Project
from . import backend
from . import operations


def execute():
    parser = ArgumentParser(
        prog='dbgear',
        description='database management tools'
    )
    parser.add_argument(
        '--project',
        default='database',
        help='please specify the folder for the project.')

    sub = parser.add_subparsers(dest='command', help='sub-command help')
    sub.add_parser('serve', help='serve help')

    apply_parser = sub.add_parser('apply', help='apply help')
    apply_parser.add_argument(
        'deploy',
        help='target deployment.')
    apply_parser.add_argument(
        'env',
        help='target environment.')
    apply_parser.add_argument(
        '--target',
        help='target table.')
    apply_parser.add_argument(
        '--all',
        choices=['drop', 'delta'],
        help='apply all tables. Specify "drop" to drop database before applying, or "delta" to apply only the changes since the last deployment.'
    )

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    logging.info(args)

    if args.command is None:
        parser.print_help()
        return

    project = Project(args.project)
    project.read_definitions()

    if args.command == 'serve':
        backend.run(project)
    if args.command == 'apply':
        if not args.all and args.target is None:
            logging.error('please specify --target or --all')
            return
        operations.apply(project, args.env, args.target, args.all, args.deploy)



from ..models.schema import Schema
from ..models.schema import Table
from ..models.schema import Field


def retrieve(folder, prefix, items, **kwargs):
    schema = Schema(prefix)

    for key, val in items.items():
        table = Table(instance=prefix, table_name=key, display_name=val)
        schema.add_table(table)

        table.fields.append(Field(
            column_name='value',
            display_name='Value',
            column_type='varchar',
            nullable=False,
            primary_key=1,
            default_value=None,
            foreign_key=None,
            comment=None
        ))
        table.fields.append(Field(
            column_name='caption',
            display_name='Caption',
            column_type='varchar',
            nullable=False,
            primary_key=None,
            default_value=None,
            foreign_key=None,
            comment=None
        ))
    return {
        prefix: schema
    }



from .base import BaseSchema


class Field(BaseSchema):
    column_name: str
    display_name: str
    column_type: str
    nullable: bool
    primary_key: int | None
    default_value: str | None
    foreign_key: str | None
    comment: str | None


class Index(BaseSchema):
    index_name: str | None
    columns: list[str]


class Table(BaseSchema):
    instance: str
    table_name: str
    display_name: str
    fields: list[Field] = []
    indexes: list[Index] = []
    # FIXME 参照元をデータとして持たせるか？


class Schema:

    def __init__(self, name):
        self.name = name
        self.tables = {}

    def __repr__(self) -> str:
        return f'{self.tables}'

    def add_table(self, table: Table) -> None:
        self.tables[table.table_name] = table

    def get_table(self, name: str) -> Table:
        return self.tables[name]

    def get_tables(self) -> dict[str, Table]:
        return self.tables


def find_field(fields: list[Field], name: str):
    field = next(filter(lambda x: x.column_name == name, fields), None)
    if field is None:
        raise RuntimeError(f'Could not find field. ({name})')
    return field



from ..dbio import engine
from ..dbio import describe

from ..models.schema import Schema
from ..models.schema import Table
from ..models.schema import Field
from ..models.schema import Index


def build_fields(conn, schema, table, primary_key):
    fields = []
    for c in describe.columns(conn, schema, table):
        fields.append(Field(
            column_name=c.COLUMN_NAME,
            display_name=c.COLUMN_NAME,
            column_type=c.COLUMN_TYPE,
            nullable=c.IS_NULLABLE == 'YES',
            primary_key=None if c.COLUMN_NAME not in primary_key else primary_key[c.COLUMN_NAME],
            default_value=c.COLUMN_DEFAULT,
            foreign_key=None,
            comment=c.COLUMN_COMMENT,
        ))
    return fields


def build_statistics(conn, schema, table):
    primary_key = {}
    indexes = {}
    for i in describe.indexes(conn, schema, table):
        if i.INDEX_NAME == 'PRIMARY':
            primary_key[i.COLUMN_NAME] = i.SEQ_IN_INDEX
        else:
            if i.INDEX_NAME in indexes:
                indexes[i.INDEX_NAME].columns.append(i.COLUMN_NAME)
            else:
                indexes[i.INDEX_NAME] = Index(
                    index_name=i.INDEX_NAME,
                    columns=[i.COLUMN_NAME],
                )
    return primary_key, indexes


def retrieve(folder, connect, mapping, **kwargs):
    schemas = {}
    with engine.get_connection(connect) as conn:
        for instance, schema in mapping.items():
            schemas[schema] = Schema(schema)

            for t in describe.tables(conn, instance):
                primary_key, indexes = build_statistics(conn, instance, t.TABLE_NAME)
                fields = build_fields(conn, instance, t.TABLE_NAME, primary_key)
                schemas[schema].add_table(Table(
                    instance=instance,
                    table_name=t.TABLE_NAME,
                    display_name=t.TABLE_NAME,
                    fields=fields,
                    indexes=list(indexes.values()),
                ))

    return schemas



from . import engine


def is_exist(conn, database):
    result = engine.select_one(conn, 'SHOW DATABASES LIKE :database', {'database': database})
    return result is not None


def create(conn, database, charset='utf8mb4', collation='utf8mb4_unicode_ci'):
    sql = f'CREATE DATABASE {database} DEFAULT CHARACTER SET {charset} COLLATE {collation}'
    engine.execute(conn, sql)


def drop(conn, database):
    sql = f'DROP DATABASE {database}'
    engine.execute(conn, sql)


# def grant(conn, database, user, host, privileges='ALL PRIVILEGES'):
#     sql = f'GRANT {privileges} ON {database}.* TO :user'
#     print(sql)
#     engine.execute(conn, sql, {'user': f'{user}@{host}'})


# def create_user(conn, user, host, password):
#     sql = f"CREATE USER IF NOT EXISTS '{user}'@'{host}' IDENTIFIED BY '{password}'"
#     engine.execute(conn, sql)



import os
from glob import glob

from ..fileio import load_model
from ..fileio import save_model
from ..fileio import make_folder
from ..fileio import delete_folder
from ..fileio import get_environ_name
from ..fileio import get_mapping_name

from .data import Mapping


def _load_mapping(folder: str, id: str) -> Mapping:
    data = load_model(
        get_mapping_name(folder, id),
        Mapping,
        id=id
    )
    if data.base is not None:
        data.parent = _load_mapping(folder, data.base)
    return data


def get(folder: str, id: str) -> Mapping:
    return _load_mapping(folder, id)


def items(folder: str) -> list[Mapping]:
    result = []
    for tmpl in glob(get_mapping_name(folder, '**')):
        id = tmpl.split('/')[-2]
        config = load_model(tmpl, Mapping, id=id)
        result.append(config)
    return sorted(result, key=lambda x: x.id)


def is_exist(folder: str, id: str) -> bool:
    return os.path.isdir(get_environ_name(folder, id))


def save(folder: str, id: str, data: Mapping) -> None:
    make_folder(get_environ_name(folder, id))
    save_model(
        get_mapping_name(folder, id),
        data,
        ['id', 'parent']
    )


def delete(folder: str, id: str) -> None:
    delete_folder(get_environ_name(folder, id))



import os
import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.responses import RedirectResponse

from .api import project
from .api import tables
from .api import environs
from .api import refs

from .models.project import Project

app = FastAPI()

app.mount('/static', StaticFiles(directory=f'{os.path.dirname(__file__)}/web', html=True), name='static')
app.include_router(project.router)
app.include_router(tables.router)
app.include_router(environs.router)
app.include_router(refs.router)


@app.get('/')
def root():
    return RedirectResponse('/static')


def run(project: Project):
    app.state.project = project
    uvicorn.run('dbgear.backend:app', port=5000, host='0.0.0.0', log_config=None)



from typing import Any
from dataclasses import asdict

from ..project import Project
from ..environ.data import Mapping
from ..schema import Table
from ..schema import find_field
from ..fileio import load_data
from .. import const

from .column import CellItem
from .data import DataModel
from .data import GridColumn
from .data import DataInfo
from . import column
from . import layout_table


def build(proj: Project, map: Mapping, dm: DataModel, table: Table) -> DataInfo:
    data = load_data(proj.folder, map.id, table.instance, table.table_name, None)
    field = find_field(table.fields, dm.y_axis)
    row_items = column.get_axis_items(proj, map, dm.settings, dm.y_axis, dm.instance)
    cells = column.make_cell_item(proj, map, dm, table)
    columns = _build_column(dm, field.display_name, row_items, cells)

    matrix = {k.value: {dm.y_axis: k.value, '_sort_key': k.caption} for k in row_items}
    for d in data:
        y_data = d[dm.y_axis]
        if y_data not in matrix:
            continue
        for cell in cells:
            matrix[y_data][cell.column_name] = d[cell.column_name]
    rows = [column.make_one_row(columns, d) for d in sorted(matrix.values(), key=lambda x: x['_sort_key'])]

    return DataInfo(
        segments=None,
        current=None,
        grid_columns=columns,
        grid_rows=rows,
        allow_line_addition_and_removal=False
    )


def _build_column(dm: DataModel, display_name: str, row_items: list[object], cells: list[CellItem]) -> list[GridColumn]:
    width = const.DEFAULT_WIDTH
    if 'width' in dm.settings[dm.y_axis]:
        width = dm.settings[dm.y_axis]['width']
    columns = []
    columns.append(column.make_grid_column(
        dm.y_axis,
        display_name,
        type=const.FIELD_TYPE_SELECTABLE,
        width=width,
        editable=False,
        items=row_items
    ))
    columns.extend([column.make_grid_column(**asdict(cell)) for cell in cells])
    return columns


def parse(proj: Project, map: Mapping, dm: DataModel, table: Table, rows: object) -> list[dict[str, Any]]:
    expdata = [
        {
            dm.y_axis: row[dm.y_axis],
            **{cell: row[cell] for cell in dm.cells}
        }
        for row in rows]
    return layout_table.parse(proj, map, dm, table, None, expdata)



from fastapi import APIRouter
from fastapi import Request

from ..models.project import project
from .dtos import Result
from .dtos import convert_to_data_filename

router = APIRouter(prefix='/tables')


@router.get('')
def get_tables(request: Request) -> Result:
    proj = project(request)
    tables = {
        ins: sorted([
            convert_to_data_filename(ins, tbl)
            for tbl in schema.get_tables().values()
        ], key=lambda x: x.table_name)
        for ins, schema in proj.schemas.items()
    }
    return Result(data=tables)


@router.get('/{instance}/{table}')
def get_table(instance: str, table: str, request: Request) -> Result:
    proj = project(request)
    schema = proj.schemas[instance]
    table = schema.get_table(table)
    return Result(data=table)



# ファイル名
FNAME_MAPPING = '_mapping.yaml'

# レイアウト
LAYOUT_TABLE = 'table'
LAYOUT_MATRIX = 'matrix'
LAYOUT_SINGLE = 'single'

# カラム定義
BIND_TYPE_FIXED = 'fixed'
BIND_TYPE_CALL = 'call'
BIND_TYPE_EXTEND = 'extend'
# カラム定義（特殊タイプ）
BIND_TYPE_REFS = 'refs'
BIND_TYPE_BLANK = ''

# callのタイプ
CALL_TYPE_UUID = 'uuid'
CALL_TYPE_ULID = 'ulid'

# FIXME セルタイプ
FIELD_TYPE_STRING = 'string'
FIELD_TYPE_SELECTABLE = 'singleSelect'
FIELD_TYPE_SYSTEM = 'system'

# 同期モード
SYNC_MODE_DROP_CREATE = 'drop_create'

# フィールド幅
DEFAULT_WIDTH = 150



import os

from ..project import Project
from ..environ.data import Mapping
from ..schema import Schema
from ..schema import Table
from ..fileio import load_model
from ..fileio import save_model
from ..fileio import save_yaml
from ..fileio import get_data_model_name
from ..fileio import get_data_dat_name
from ..datagrid import grid
from ..datagrid.data import DataInfo
from ..datagrid.data import DataModel

from . import mapping


def get(proj: Project, map: Mapping, ins: str, tbl: str, seg: str | None) -> tuple[DataModel, Table, DataInfo]:
    dm = load_model(
        get_data_model_name(proj.folder, map.id, ins, tbl),
        DataModel,
        id=map.id,
        instance=ins,
        table_name=tbl
    )
    table = proj.schemas[ins].get_table(tbl)
    info = grid.build(proj, map, dm, table, seg)

    return (dm, table, info)


def get_row(proj: Project, map: Mapping, ins: str, tbl: str) -> dict[str, object]:
    dm = load_model(
        get_data_model_name(proj.folder, map.id, ins, tbl),
        DataModel,
        id=map.id,
        instance=ins,
        table_name=tbl
    )
    table = proj.schemas[ins].get_table(tbl)
    row = grid.build_one_row(proj, map, dm, table)
    return row


def get_description(proj: Project, id: str, ins: str, tbl: str) -> str:
    return load_model(
        get_data_model_name(proj.folder, id, ins, tbl),
        DataModel,
        id=id,
        instance=ins,
        table_name=tbl
    ).description


def items(schemas: dict[str, Schema], folder: str, id: str, *, exist: bool = True) -> list[tuple[str, Table]]:
    config = mapping.get(folder, id)
    tables = []
    for ins in config.instances:
        if ins not in schemas:
            # 元の定義から消えているため無視する。
            continue
        tables.extend([
            (ins, tbl)
            for tbl in schemas[ins].get_tables().values()
            if is_exist(folder, id, ins, tbl.table_name) == exist
        ])
    return sorted(tables, key=lambda x: f'{x[0]}@{x[1].table_name}')


def is_exist(folder: str, id: str, ins: str, tbl: str) -> bool:
    return os.path.isfile(get_data_model_name(folder, id, ins, tbl))


def save(proj: Project, map: Mapping, ins: str, tbl: str, data: DataModel) -> None:
    save_model(
        get_data_model_name(proj.folder, map.id, ins, tbl),
        data,
        ['id', 'instance', 'table_name']
    )


def save_data(proj: Project, map: Mapping, ins: str, tbl: str, seg: str | None, rows: object) -> None:
    dm = load_model(
        get_data_model_name(proj.folder, map.id, ins, tbl),
        DataModel,
        id=map.id,
        instance=ins,
        table_name=tbl
    )
    table = proj.schemas[ins].get_table(tbl)
    data = grid.parse(proj, map, dm, table, seg, rows)
    save_yaml(
        get_data_dat_name(proj.folder, map.id, ins, tbl, seg),
        data
    )



from . import engine


def tables(conn, schema):
    sql = f"SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = '{schema}'"
    return engine.select_all(conn, sql)


def columns(conn, schema, table):
    sql = f'''
        SELECT
        COLUMN_NAME
        , COLUMN_TYPE
        , IS_NULLABLE
        , COLUMN_KEY
        , COLUMN_DEFAULT
        , COLUMN_COMMENT
        FROM information_schema.columns
        WHERE table_schema = '{schema}' AND table_name = '{table}'
        ORDER BY ORDINAL_POSITION
    '''
    return engine.select_all(conn, sql)


def indexes(conn, schema, table):
    sql = f'''
        SELECT
        INDEX_NAME
        , NON_UNIQUE
        , COLUMN_NAME
        , SEQ_IN_INDEX
        FROM information_schema.statistics
        WHERE table_schema = '{schema}' AND table_name = '{table}'
        ORDER BY INDEX_NAME, SEQ_IN_INDEX
    '''
    return engine.select_all(conn, sql)



from sqlalchemy import create_engine
from sqlalchemy import text


def _get_engine(conn):
    return create_engine(conn, echo=False)


def get_connection(conn):
    engine = _get_engine(conn)
    return engine.connect()


def execute(conn, sql, params=None):
    return conn.execute(text(sql), params)


def select_all(conn, sql, params=None):
    return execute(conn, sql, params).mappings().fetchall()


def select_one(conn, sql, params=None):
    return execute(conn, sql, params).mappings().fetchone()


def commit(conn):
    conn.commit()



from typing import Any
from dataclasses import dataclass
from uuid import uuid4
from ulid import ULID

from ..project import Project
from ..environ.data import Mapping
from ..schema import Table
from ..schema import find_field
from ..fileio import load_all_data
from ..fileio import load_model
from ..fileio import get_data_model_name
from .. import const

from .data import DataModel
from .data import ListItem
from .data import GridColumn


def make_grid_column(
        column_name: str,
        display_name: str,
        *,
        type: str = const.FIELD_TYPE_STRING,
        width: int = const.DEFAULT_WIDTH,
        editable: bool = True,
        hide: bool = False,
        items: list[ListItem] | None = None,
        fixed_value: str | None = None,
        call_value: str | None = None):
    '''
    GridColumnを生成する。
    '''
    return GridColumn(
        field=column_name,
        type=type,
        header_name=display_name,
        width=width,
        editable=editable,
        hide=hide,
        items=items,
        fixed_value=fixed_value,
        call_value=call_value,
    )


def load_for_select_items(folder: str, id: str, ins: str, ref: str) -> list[ListItem]:
    '''
    マスタデータをロードして、選択肢用データを生成する。
    '''
    items = load_all_data(folder, id, ins, ref)
    if items is not None:
        # データが見つかった場合には、DataModelをロードし、
        # caption, valueのデータを生成する
        dm: DataModel = load_model(
            get_data_model_name(folder, id, ins, ref),
            DataModel,
            id=id,
            instance=ins,
            table_name=ref
        )
        if dm.caption is not None and dm.value is not None:
            items = [ListItem(
                caption=item[dm.caption],
                value=item[dm.value]
            ) for item in items]
        else:
            raise RuntimeError('Invalid Data Model for Master Data.')
    return items


@dataclass
class CellItem:
    column_name: str
    display_name: str
    type: str = const.FIELD_TYPE_STRING
    width: int = const.DEFAULT_WIDTH
    editable: bool = True
    items: list[ListItem] | None = None
    fixed_value: str | None = None
    call_value: str | None = None


def make_cell_item(proj: Project, map: Mapping, dm: DataModel, table: Table) -> list[CellItem]:
    '''
    セルの設定を生成する。
    '''
    result = []
    for cell in dm.cells:
        field = find_field(table.fields, cell)
        if cell in dm.settings:
            setting = dm.settings[cell]
            if setting['type'] == const.BIND_TYPE_BLANK:
                result.append(CellItem(
                    column_name=field.column_name,
                    display_name=field.display_name,
                    width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH
                ))
            elif setting['type'] == const.BIND_TYPE_REFS:
                id = setting['id']
                ins = setting['instance']
                tbl = setting['table']
                items = load_for_select_items(proj.folder, id, ins, tbl)

                result.append(CellItem(
                    column_name=field.column_name,
                    display_name=field.display_name,
                    type=const.FIELD_TYPE_SELECTABLE,
                    width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
                    editable=True,
                    items=items
                ))
            else:
                bind = proj.bindings[setting['type']]
                if bind.type == const.BIND_TYPE_FIXED:
                    result.append(CellItem(
                        column_name=field.column_name,
                        display_name=field.display_name,
                        width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
                        editable=False,
                        fixed_value=bind.value
                    ))
                elif bind.type == const.BIND_TYPE_CALL:
                    result.append(CellItem(
                        column_name=field.column_name,
                        display_name=field.display_name,
                        width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
                        editable=False,
                        call_value=bind.value
                    ))
                elif bind.type == const.BIND_TYPE_EXTEND:
                    result.append(CellItem(
                        column_name=field.column_name,
                        display_name=field.display_name,
                        width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
                        editable=True
                    ))
        else:
            result.append(CellItem(
                column_name=field.column_name,
                display_name=field.display_name
            ))
    return result


def exclude_names(items: list[tuple[str, Any]]) -> dict[str, Any]:
    '''
    column_name, display_nameを除外した辞書を生成する。
    '''
    return {key: value for key, value in items if key not in ['column_name', 'display_name']}


def adjust_column_value(col: GridColumn, value: Any, fixed: bool = False) -> Any:
    '''
    valueの中にフィールドのキーがなかったら、値を補完する。
    fixedがTrueの場合には、fixed_valueを優先する。
    '''
    if fixed and col.fixed_value is not None:
        return col.fixed_value
    if col.field in value:
        return value[col.field]
    if col.fixed_value is not None:
        return col.fixed_value
    if col.call_value is not None:
        if col.call_value == const.CALL_TYPE_UUID:
            return str(uuid4())
        if col.call_value == const.CALL_TYPE_ULID:
            return str(ULID())
    return ''


def make_one_row(columns: list[GridColumn], data: Any, *, need_id: bool = True, fixed: bool = False, segment: tuple = (False, None, None)) -> dict[str, Any]:
    '''
    1行分のデータを生成する。
    '''
    row = {
        col.field: adjust_column_value(col, data, fixed)
        for col in columns
    }

    # セグメントがある場合には、セグメントの値を設定する
    if segment[0]:
        row[segment[1]] = segment[2]

    # idが必要な場合には、idを生成する
    if need_id and 'id' not in row:
        row['id'] = str(uuid4())

    return row


def get_axis_items(proj: Project, map: Mapping, settings: dict[str, object], axis: str, ins: str) -> list[ListItem]:
    items = None
    setting = settings[axis]
    if setting['type'] == const.BIND_TYPE_REFS:
        id = setting['id']
        ins = setting['instance']
        tbl = setting['table']
        items = load_for_select_items(proj.folder, id, ins, tbl)
    if items is None:
        raise RuntimeError(f'No Data for axis. ({ins}@{axis})')
    return items



from typing import Any

from ..project import Project
from ..environ.data import Mapping
from ..schema import Table
from ..schema import Field
from ..fileio import load_data
from .. import const

from .data import DataModel
from .data import ListItem
from .data import GridColumn
from .data import DataInfo
from . import column


def build(proj: Project, map: Mapping, dm: DataModel, table: Table, segment: str | None) -> DataInfo:
    # フィールドを設定に従って展開する。ForeignKeyなども設定に含まれるものとする。
    columns, seg_info = build_columns(proj, map, dm, table)
    if seg_info is None:
        data = load_data(proj.folder, map.id, table.instance, table.table_name, None)
        rows = [column.make_one_row(columns, d) for d in data]
        return DataInfo(
            segments=None,
            current=None,
            grid_columns=columns,
            grid_rows=rows,
            allow_line_addition_and_removal=True
        )

    if segment is None:
        segment = seg_info[1][0].value

    data = load_data(proj.folder, map.id, table.instance, table.table_name, segment)
    rows = [column.make_one_row(columns, d) for d in data]

    return DataInfo(
        segments=seg_info[1],
        current=segment,
        grid_columns=columns,
        grid_rows=rows,
        allow_line_addition_and_removal=True
    )


def build_one_row(proj: Project, map: Mapping, dm: DataModel, table: Table) -> dict[str, Any]:
    # ここでは画面表示用のデータ生成のため、セグメントは不要。
    columns, _ = build_columns(proj, map, dm, table)
    return column.make_one_row(columns, {})


def build_columns(proj: Project, map: Mapping, dm: DataModel, table: Table) -> tuple[list[GridColumn], tuple[str, list[ListItem]] | None]:
    columns = []
    items = None
    for field in table.fields:
        if field.column_name == dm.segment:
            # セグメントは表示しないせずに、セグメントリストを作成する。
            setting = dm.settings[field.column_name]
            # 仮にrefs以外が設定されていた場合にはidやinstanceなどがないので、エラーとなる。
            assert setting['type'] == const.BIND_TYPE_REFS
            id = setting['id']
            ins = setting['instance']
            tbl = setting['table']
            items = column.load_for_select_items(proj.folder, id, ins, tbl)
            if items is None:
                raise RuntimeError('Segment items not found.')
            continue
        grid_column = _make_grid_column_from_setting(proj, map, dm, field)
        columns.append(grid_column)
    if items is None:
        return columns, None
    return columns, (dm.segment, items)


def _make_grid_column_from_setting(proj: Project, map: Mapping, dm: DataModel, field: Field):
    if field.column_name not in dm.settings:
        return column.make_grid_column(
            field.column_name,
            field.display_name)

    setting = dm.settings[field.column_name]
    if setting['type'] == const.BIND_TYPE_BLANK:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH)

    if setting['type'] == const.BIND_TYPE_REFS:
        # 選択肢は外部データを参照する以外は選択型と同じ
        id = setting['id']
        ins = setting['instance']
        tbl = setting['table']
        items = column.load_for_select_items(proj.folder, id, ins, tbl)
        if items is None:
            # データが見つからなければ空の配列とする
            items = []
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            type=const.FIELD_TYPE_SELECTABLE,
            width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
            items=items)

    bind = proj.bindings[setting['type']]
    if bind.type == const.BIND_TYPE_FIXED:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
            editable=False,
            hide=True,
            fixed_value=bind.value)
    if bind.type == const.BIND_TYPE_CALL:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
            editable=False,
            call_value=bind.value)
    if bind.type == const.BIND_TYPE_EXTEND:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            width=setting['width'] if 'width' in setting else const.DEFAULT_WIDTH,
            editable=True)
    raise RuntimeError('Unknown Data Type')


def parse(proj: Project, map: Mapping, dm: DataModel, table: Table, segment: str | None, rows: object) -> list[dict[str, Any]]:
    # idカラムを除いて返却する。
    columns, seg_info = build_columns(proj, map, dm, table)
    if seg_info is None:
        data = [column.make_one_row(columns, d, need_id=False) for d in rows]
        return data
    # seg_infoがある場合には、セグメントカラムを作成する。
    assert segment is not None
    data = [
        column.make_one_row(
            columns,
            d,
            need_id=False,
            segment=(True, seg_info[0], segment)
        ) for d in rows
    ]
    return data



from ..models.base import BaseSchema
from ..models.schema import Table
from ..models.project import Binding
from ..models.environ.data import Mapping
from ..models.datagrid.data import DataModel
from ..models.datagrid.data import DataInfo

# リクエスト用DTO


class NewMapping(BaseSchema):
    group: str
    base: str | None
    name: str
    instances: list = []
    description: str = ''
    deployment: bool


def convert_to_mapping(id: str, data: NewMapping) -> Mapping:
    return Mapping(
        id=id,
        **data.model_dump()
    )


class NewDataModel(BaseSchema):
    description: str
    layout: str
    settings: dict[str, object]
    sync_mode: str
    value: str | None = None
    caption: str | None = None
    segment: str | None = None
    x_axis: str | None = None
    y_axis: str | None = None
    cells: list[str] | None = None


def convert_to_data_model(
        data: NewDataModel,
        id: str = '',
        instance: str = '',
        table_name: str = '') -> DataModel:
    return DataModel(
        id=id,
        instance=instance,
        table_name=table_name,
        **data.model_dump()
    )


class ImportSQL(BaseSchema):
    host: str
    sql: str


# レスポンス用DTO


class Result(BaseSchema):
    status: str = 'OK'
    message: str | None = None
    data: object | list | None = None


class ProjectInfo(BaseSchema):
    project_name: str
    description: str
    bindings: dict[str, Binding]
    rules: dict[str, str]
    instances: list[str]
    api_key: str | None = None


class DataFilename(BaseSchema):
    id: str | None
    id_name: str | None
    instance: str
    table_name: str
    display_name: str | None
    description: str | None


def convert_to_data_filename(instance: str, tbl: Table, description: str | None = None, id: str | None = None, id_name: str | None = None):
    return DataFilename(
        id=id,
        id_name=id_name,
        instance=instance,
        table_name=tbl.table_name,
        display_name=tbl.display_name,
        description=description
    )


class Data(BaseSchema):
    model: DataModel
    info: DataInfo
    table: Table


class MappingTree(BaseSchema):
    name: str | None
    children: list[Mapping]



from typing import Any
from ..base import BaseSchema


class DataModel(BaseSchema):
    id: str
    instance: str
    table_name: str
    description: str = ''
    layout: str
    settings: dict[str, object]
    sync_mode: str
    value: str | None = None
    caption: str | None = None
    segment: str | None = None
    x_axis: str | None = None
    y_axis: str | None = None
    cells: list[str] | None = None


class ListItem(BaseSchema):
    caption: str
    value: str


class GridColumn(BaseSchema):
    field: str
    type: str
    header_name: str
    width: int
    editable: bool
    hide: bool = False
    items: list[ListItem] | None = None
    fixed_value: str | None = None
    call_value: str | None = None


class DataInfo(BaseSchema):
    segments: list[ListItem] | None
    current: str | None
    grid_columns: list[GridColumn]
    grid_rows: list[dict[str, Any]]
    allow_line_addition_and_removal: bool



import os
import shutil
import yaml
from typing import Any
from glob import glob

from . import const


def get_environ_name(folder: str, id: str):
    return f'{folder}/{id}'


def get_mapping_name(folder: str, id: str):
    return f'{folder}/{id}/{const.FNAME_MAPPING}'


def get_data_model_name(folder: str, id: str, instance: str, table: str) -> str:
    return f'{folder}/{id}/{instance}@{table}.yaml'


def is_exist_data_model(folder: str, id: str, instance: str, table: str) -> bool:
    return os.path.isfile(get_data_model_name(folder, id, instance, table))


def get_data_dat_name(folder: str, id: str, instance: str, table: str, seg: str | None) -> str:
    if seg is not None:
        return f'{folder}/{id}/{instance}@{table}#{seg}.dat'
    return f'{folder}/{id}/{instance}@{table}.dat'


def save_model(fname: str, model: Any, exclude: list[str] = None) -> None:
    with open(fname, 'w', encoding='utf-8') as f:
        yaml.dump(model.model_dump(exclude=exclude), f, indent=2, allow_unicode=True)


def save_yaml(fname: str, data: object) -> None:
    with open(fname, 'w', encoding='utf-8') as f:
        yaml.dump(data, f, indent=2, allow_unicode=True)


def load_model(fname: str, clazz: Any, **kwargs) -> Any:
    return clazz(**load_yaml(fname), **kwargs)


def load_yaml(fname: str) -> Any:
    with open(fname, 'r', encoding='utf-8') as f:
        data = yaml.safe_load(f)
    return data


def _is_exist_raw_data(folder: str, id: str, ins: str, tbl: str, seg: str | None) -> bool:
    return os.path.isfile(get_data_dat_name(folder, id, ins, tbl, seg))


def load_data(folder: str, id: str, ins: str, tbl: str, seg: str | None) -> Any:
    data = []
    if _is_exist_raw_data(folder, id, ins, tbl, seg):
        data = load_yaml(get_data_dat_name(folder, id, ins, tbl, seg))
    return data


def load_all_data(folder: str, id: str, ins: str, tbl: str) -> Any:
    # セグメントで分かれていても、一つのデータとしてロードする
    data = None
    if _is_exist_raw_data(folder, id, ins, tbl, None):
        return load_yaml(get_data_dat_name(folder, id, ins, tbl, None))

    for f in glob(get_data_dat_name(folder, id, ins, tbl, '*')):
        if data is None:
            data = []
        data.extend(load_yaml(f))

    return data


def make_folder(dirname: str) -> None:
    if not os.path.isdir(dirname):
        os.mkdir(dirname)


def delete_folder(dirname: str) -> None:
    shutil.rmtree(dirname)



from logging import getLogger
from importlib import import_module

from .base import BaseSchema
from .schema import Schema
from .fileio import load_yaml


class Binding(BaseSchema):
    type: str
    value: str


class Project:
    logger = getLogger('project')

    def __init__(self, folder: str):
        self.folder = folder

        data = load_yaml(f'{folder}/project.yaml')
        self.project_name = data['project_name']
        self.description = data['description']
        self._rules = data['rules']
        self._definitions = data['definitions']
        self._bindings = {k: Binding(**v) for k, v in data['bindings'].items()}
        self._schemas = {}
        self.deployments = data['deployments']

    @property
    def bindings(self) -> dict[str, Binding]:
        return self._bindings

    @property
    def rules(self) -> dict[str, str]:
        return self._rules

    @property
    def schemas(self) -> dict[str, Schema]:
        return self._schemas

    @property
    def instances(self) -> list[str]:
        return list(self._schemas.keys())

    def read_definitions(self) -> None:
        for items in self._definitions:
            self.logger.info(f"definition: {items['type']}")
            module = import_module(f'.{items["type"]}', 'dbgear.definitions')
            result = module.retrieve(self.folder, **items)

            self._schemas.update(result)


def project(request) -> Project:
    return request.app.state.project



from typing import Any

from ..project import Project
from ..environ.data import Mapping
from ..schema import Table
from .. import const

from .data import DataModel
from .data import DataInfo
from . import layout_table
from . import layout_matrix
from . import layout_single


def build(proj: Project, map: Mapping, dm: DataModel, table: Table, segment: str | None) -> DataInfo:
    if dm.layout == const.LAYOUT_TABLE:
        return layout_table.build(proj, map, dm, table, segment)
    if dm.layout == const.LAYOUT_MATRIX:
        return layout_matrix.build(proj, map, dm, table)
    if dm.layout == const.LAYOUT_SINGLE:
        return layout_single.build(proj, map, dm, table)
    raise RuntimeError(f'Unknown layout. {dm.layout}')


def build_one_row(proj: Project, map: Mapping, dm: DataModel, table: Table) -> dict[str, Any]:
    if dm.layout == const.LAYOUT_TABLE:
        return layout_table.build_one_row(proj, map, dm, table)
    # このレイアウトでは行を作成できないの例外を発生させる
    raise RuntimeError('Cannot build one row at this layout.')


def parse(proj: Project, map: Mapping, dm: DataModel, table: Table, segment: str | None, data: Any) -> list[dict[str, Any]]:
    if dm.layout == const.LAYOUT_TABLE:
        return layout_table.parse(proj, map, dm, table, segment, data)
    if dm.layout == const.LAYOUT_MATRIX:
        return layout_matrix.parse(proj, map, dm, table, data)
    if dm.layout == const.LAYOUT_SINGLE:
        return layout_single.parse(proj, map, dm, table, data)
    raise RuntimeError(f'Unknown layout. {dm.layout}')



import importlib

from logging import getLogger
from datetime import datetime
from .dbio import engine
from .dbio import database
from .dbio import table

from .models.project import Project
from .models.environ import mapping
from .models.environ.data import Mapping
from .models.datagrid.data import DataModel
from .models.fileio import load_all_data
from .models.fileio import get_data_model_name
from .models.fileio import is_exist_data_model
from .models.fileio import load_model
from .models import const

logger = getLogger(__name__)


def _load_for_entry(proj: Project, map: Mapping, ins: str, tbl: str, dm: DataModel):
    # 自身のデータをロードし、存在しなければ親を再帰的に遡ってロードする。
    items = load_all_data(proj.folder, map.id, ins, tbl)
    if items is None:
        if map.parent is not None:
            return _load_for_entry(proj, map.parent, ins, tbl, dm)
    # extendの処理
        return None
    for item in items:
        for col, val in item.items():
            if col in dm.settings:
                typ = dm.settings[col]['type']
                if typ in [const.BIND_TYPE_BLANK, const.BIND_TYPE_REFS]:
                    continue
                bind = proj.bindings[typ]
                if bind is None:
                    continue
                if bind.type == const.BIND_TYPE_EXTEND:
                    module = importlib.import_module(bind.value)
                    result = module.convert(proj, map, ins, tbl, dm, *val.split(','))
                    item[col] = result
    return items


def _load_data_model(proj: Project, map: Mapping, ins: str, tbl: str) -> bool:
    if is_exist_data_model(proj.folder, map.id, ins, tbl) is False:
        if map.parent is None:
            return None
        return _load_data_model(proj, map.parent, ins, tbl)

    dm = load_model(
        get_data_model_name(proj.folder, map.id, ins, tbl),
        DataModel,
        id=map.id,
        instance=ins,
        table_name=tbl
    )
    return dm


class Operation:

    @staticmethod
    def get_instance(folder: str, env: str, deploy: str):
        project = Project(folder)
        project.read_definitions()
        return Operation(project, env, deploy)

    def __init__(self, project: Project, env: str, deploy: str):
        self.project = project
        self.map = mapping.get(project.folder, env)
        self.conn = engine.get_connection(project.deployments[deploy])
        self.ymd = datetime.strftime(datetime.now(), '%Y%m%d%H%M%S')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def create_database(self, all: str):
        if all == 'drop':
            logger.info(f'database {self.map.id}')
            if database.is_exist(self.conn, self.map.id):
                database.drop(self.conn, self.map.id)
            database.create(self.conn, self.map.id)
        else:
            # 差分更新または個別指定の場合で、データベースが存在しない場合は作成する。
            if not database.is_exist(self.conn, self.map.id):
                logger.info(f'database {self.map.id} was created.')
                database.create(self.conn, self.map.id)

    def create_table(self, all: str, target: str):
        for ins in self.map.instances:
            schema = self.project.schemas[ins]
            for tbl in schema.tables.values():
                if not all and target != tbl.table_name:
                    continue
                # テーブルが存在しない場合は作成する。
                if table.is_exist(self.conn, self.map.id, tbl) is False:
                    logger.info(f'table {self.map.id}.{tbl.table_name} was created.')
                    table.create(self.conn, self.map.id, tbl)
                    continue
                # データのバックアップ
                logger.info(f'backup {self.map.id}.{tbl.table_name}')
                table.backup(self.conn, self.map.id, tbl, self.ymd)
                # テーブルの再作成
                logger.info(f'drop & create table {self.map.id}.{tbl.table_name}')
                table.drop(self.conn, self.map.id, tbl)
                table.create(self.conn, self.map.id, tbl)

    def insert_data(self, all: bool, target: str):
        # データ投入
        for ins in self.map.instances:
            schema = self.project.schemas[ins]
            for tbl in schema.tables.values():
                if not all and target != tbl.table_name:
                    continue
                dm = _load_data_model(self.project, self.map, ins, tbl.table_name)
                items = _load_for_entry(self.project, self.map, ins, tbl.table_name, dm)
                if items is not None:
                    logger.info(f'insert {self.map.id}.{tbl.table_name}')
                    table.insert(self.conn, self.map.id, tbl, items)
                if dm is not None and dm.sync_mode == 'update_diff':
                    # 新規作成時にはバックアップはない
                    if table.is_exist_backup(self.conn, self.map.id, tbl, self.ymd):
                        # バックアップからデータを復元
                        # create_tableの中で再作成されているため
                        logger.info(f'restore {self.map.id}.{tbl.table_name}')
                        table.restore(self.conn, self.map.id, tbl, self.ymd)
                engine.commit(self.conn)

    def reset_all(self):
        # ユニットテストなど一括して再作成を行う場合に使用する。
        self.create_database('drop')
        self.create_table('drop', None)

    def require(self, instance: str, table_name: str):
        # ユニットテストなどで指定したデータを挿入する。
        schema = self.project.schemas[instance]
        tbl = schema.get_table(table_name)
        dm = _load_data_model(self.project, self.map, instance, table_name)
        items = _load_for_entry(self.project, self.map, instance, table_name, dm)
        if items is not None:
            logger.info(f'insert {self.map.id}.{table_name}')
            table.insert(self.conn, self.map.id, tbl, items)


def apply(project, env: str, target: str, all: str, deploy: str):
    with Operation(project, env, deploy) as op:
        op.create_database(all)
        op.create_table(all, target)
        op.insert_data(all, target)



from ..project import Project
from ..environ.data import Mapping
from ..datagrid.data import DataModel
from ..datagrid import layout_table
from ..datagrid import column
from ..fileio import load_model
from ..fileio import save_yaml
from ..fileio import get_data_model_name
from ..fileio import get_data_dat_name

from ...dbio import engine


def execute(proj: Project, map: Mapping, ins: str, tbl: str, seg: str | None, host: str, sql: str):
    dm = load_model(
        get_data_model_name(proj.folder, map.id, ins, tbl),
        DataModel,
        id=map.id,
        instance=ins,
        table_name=tbl
    )
    columns, seg_info = layout_table.build_columns(proj, map, dm, proj.schemas[ins].get_table(tbl))
    with engine.get_connection(proj.deployments[host]) as conn:
        data = engine.select_all(conn, sql)

    if seg_info is None:
        rows = [column.make_one_row(columns, d, need_id=False, fixed=True) for d in data]
    else:
        # seg_infoがある場合には、segの引数は指定されている必要がある。
        assert seg is not None
        rows = [column.make_one_row(columns, d, need_id=False, fixed=True, segment=(True, seg_info[0], seg)) for d in data]

    save_yaml(
        get_data_dat_name(proj.folder, map.id, ins, tbl, seg),
        rows
    )



import re
from glob import glob

from fastapi import APIRouter
from fastapi import Request

from ..models.project import project
from ..models.environ.data import Mapping
from ..models.fileio import load_model
from ..models.fileio import get_mapping_name
from ..models.fileio import get_environ_name
from .dtos import Result
from .dtos import DataFilename

router = APIRouter(prefix='/refs')


@router.get('')
def get_referencable(request: Request) -> Result:
    proj = project(request)

    # データを横断して検索して一覧を生成する
    data = []
    for tmpl in glob(get_mapping_name(proj.folder, '**')):
        id = tmpl.split('/')[-2]
        map = load_model(tmpl, Mapping, id=id)
        for dat in glob(get_environ_name(proj.folder, id) + '/*.dat'):
            name = dat.split('/')[-1]
            if '#' in name:
                (ins, tbl, _, _) = re.split('[.@#]', name)
            else:
                (ins, tbl, _) = re.split('[.@]', name)
            table = proj.schemas[ins].get_table(tbl)
            data.append(DataFilename(
                id=id,
                instance=ins,
                table_name=tbl,
                display_name=table.display_name,
                description=None,
                id_name=map.name
            ))

    return Result(data=data)



from fastapi import APIRouter
from fastapi import Request
from fastapi import Body

from ..models.project import project
from ..models.environ import mapping
from ..models.environ import entity
from ..models.environ import import_sql

from .dtos import Result
from .dtos import Data
from .dtos import NewMapping
from .dtos import NewDataModel
from .dtos import ImportSQL
from .dtos import MappingTree
from .dtos import convert_to_data_filename
from .dtos import convert_to_mapping
from .dtos import convert_to_data_model

router = APIRouter(prefix='/environs')


@router.get('')
def get_mappings(request: Request):
    proj = project(request)
    maps = mapping.items(proj.folder)

    groupTree = {}
    for map in maps:
        if map.group not in groupTree:
            groupTree[map.group] = MappingTree(name=map.group, children=[])
        groupTree[map.group].children.append(map)

    return Result(data=[*groupTree.values()])


@router.post('/{id}')
def create_mapping(id: str, data: NewMapping, request: Request) -> Result:
    proj = project(request)
    if mapping.is_exist(proj.folder, id):
        return Result(
            status='ERROR',
            message='ERROR_EXIST_MAPPING'
        )
    mapping.save(proj.folder, id, convert_to_mapping(id, data))
    return Result()


@router.get('/{id}/tables')
def get_tables(id: str, request: Request):
    '''
    既に存在するテーブル一覧。
    '''
    proj = project(request)
    tables = [
        convert_to_data_filename(
            ins,
            tbl,
            entity.get_description(proj, id, ins, tbl.table_name)
        )
        for ins, tbl in entity.items(proj.schemas, proj.folder, id)
    ]
    return Result(data=tables)


@router.get('/{id}/init')
def get_not_exist_tables(id: str, request: Request):
    '''
    新規作成時のテーブル一覧。既に存在するものは除く
    '''
    proj = project(request)
    tables = [
        convert_to_data_filename(ins, tbl)
        for ins, tbl in entity.items(proj.schemas, proj.folder, id, exist=False)
    ]
    return Result(data=tables)


@router.get('/{id}/tables/{instance}/{table}')
def get_table(id: str, instance: str, table: str, request: Request, segment: str | None = None):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    dm, tbl, info = entity.get(proj, map, instance, table, segment)
    data = Data(
        model=dm,
        info=info,
        table=tbl
    )
    return Result(data=data)


@router.get('/{id}/tables/{instance}/{table}/row')
def get_row(id: str, instance: str, table: str, request: Request):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    row = entity.get_row(proj, map, instance, table)
    return Result(data=row)


@router.post('/{id}/tables/{instance}/{table}')
def create_data_model(id: str, instance: str, table: str, data: NewDataModel, request: Request):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    entity.save(proj, map, instance, table, convert_to_data_model(data))
    return Result()


@router.put('/{id}/tables/{instance}/{table}')
def update_data(id: str, instance: str, table: str, request: Request, body=Body(...), segment: str | None = None):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    entity.save_data(proj, map, instance, table, segment, body)
    return Result()


@router.post('/{id}/tables/{instance}/{table}/import')
def import_data(id: str, instance: str, table: str, imp: ImportSQL, request: Request, segment: str | None = None):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    import_sql.execute(proj, map, instance, table, segment, imp.host, imp.sql)
    return Result()
